@@ -7,9 +7,9 @@ Provides an implementation of the Pelix HTTP service based on the standard
 Python library.
 
 :author: Thomas Calmant
-:copyright: Copyright 2012, isandlaTech
+:copyright: Copyright 2013, isandlaTech
 :license: GPLv3
-:version: 0.1
+:version: 0.1.1
 :status: Alpha
 
 ..
@@ -30,7 +30,9 @@ Python library.
     along with iPOPO. If not, see <http://www.gnu.org/licenses/>.
 """
 
-__version__ = (0, 1, 0)
+# Module version
+__version_info__ = (0, 1, 1)
+__version__ = ".".join(map(str, __version_info__))
 
 # Documentation strings format
 __docformat__ = "restructuredtext en"
@@ -38,7 +40,7 @@ __docformat__ = "restructuredtext en"
 # ------------------------------------------------------------------------------
 
 from pelix.ipopo.decorators import ComponentFactory, Provides, Validate, \
-    Invalidate, Property, Instantiate, Requires, Bind, Unbind
+    Invalidate, Property, Requires, Bind, Unbind
 
 import pelix.ipopo.constants as constants
 import pelix.framework as pelix
@@ -70,6 +72,7 @@ else:
 import pelix.http as http
 
 HTTP_SERVICE_COMPONENT_FACTORY = "pelix.http.service.basic.factory"
+""" Name of the HTTP service component factory """
 
 # ------------------------------------------------------------------------------
 
@@ -251,11 +254,11 @@ class _RequestHandler(BaseHTTPRequestHandler):
         self._service.log(logging.ERROR, message, *args, **kwargs)
 
 
-    def log_request(self, message, *args, **kwargs):
+    def log_request(self, code='-', size='-'):
         """
-        Log a request to the server
+        Logs a request to the server
         """
-        self._service.log(logging.DEBUG, message, *args, **kwargs)
+        self._service.log(logging.DEBUG, '"%s" %s', self.requestline, code)
 
 
     def send_no_servlet_response(self):
@@ -336,7 +339,7 @@ class _HttpServerFamily(ThreadingMixIn, HTTPServer):
 @Property("_port", http.HTTP_SERVICE_PORT, 8080)
 @Property("_instance_name", constants.IPOPO_INSTANCE_NAME)
 @Property("_logger_name", "pelix.http.logger.name", "")
-@Property("_logger_level", "pelix.http.logger.level", logging.ERROR)
+@Property("_logger_level", "pelix.http.logger.level", None)
 class HttpService(object):
     """
     Basic HTTP service component
@@ -634,7 +637,12 @@ class HttpService(object):
                 self._logger_name = self._instance_name
 
             self._logger = logging.getLogger(self._logger_name)
-            self._logger.level = int(self._logger_level)
+
+            if self._logger_level is None:
+                self._logger.level = logging.INFO
+
+            else:
+                self._logger.level = int(self._logger_level)
 
         self.log(logging.INFO, "Starting HTTP server: [%s]:%d ...",
                  self._address, self._port)
