@@ -4,9 +4,9 @@
 Core iPOPO implementation
 
 :author: Thomas Calmant
-:copyright: Copyright 2012, isandlaTech
+:copyright: Copyright 2013, isandlaTech
 :license: GPLv3
-:version: 0.4
+:version: 0.5
 :status: Alpha
 
 ..
@@ -27,7 +27,9 @@ Core iPOPO implementation
     along with iPOPO. If not, see <http://www.gnu.org/licenses/>.
 """
 
-__version__ = (0, 4, 0)
+# Module version
+__version_info__ = (0, 5, 0)
+__version__ = ".".join(map(str, __version_info__))
 
 # Documentation strings format
 __docformat__ = "restructuredtext en"
@@ -36,8 +38,7 @@ __docformat__ = "restructuredtext en"
 
 from pelix.framework import BundleContext, ServiceEvent, BundleEvent, \
     Bundle, BundleException
-from pelix.utilities import remove_all_occurrences, SynchronizedClassMethod, \
-    add_listener, remove_listener, is_string
+from pelix.utilities import add_listener, remove_listener, is_string
 
 import pelix.ipopo.constants as constants
 import pelix.framework as pelix
@@ -328,11 +329,12 @@ class _RuntimeDependency(object):
         raise NotImplementedError
 
 
-    def on_service_modify(self, svc_ref):
+    def on_service_modify(self, svc_ref, old_properties):
         """
         Called when a service has been registered in the framework
         
         :param svc_ref: A service reference
+        :param old_properties: Previous properties values
         """
         raise NotImplementedError
 
@@ -341,12 +343,15 @@ class _RuntimeDependency(object):
         """
         Called by the framework when a service event occurs
         """
-        if not self._ipopo_instance.check_event(event):
-            # We've been told to ignore this event
+        if self._ipopo_instance is None \
+        or not self._ipopo_instance.check_event(event):
+            # stop() and clean() may have been called after we have been put
+            # inside a listener list copy...
+            # or we've been told to ignore this event
             return
 
         # Call sub-methods
-        kind = event.get_type()
+        kind = event.get_kind()
         svc_ref = event.get_service_reference()
 
         if kind == ServiceEvent.REGISTERED:
@@ -360,7 +365,7 @@ class _RuntimeDependency(object):
 
         elif kind == ServiceEvent.MODIFIED:
             # Modified properties (can be a new injection)
-            self.on_service_modify(svc_ref)
+            self.on_service_modify(svc_ref, event.get_previous_properties())
 
 
     def start(self):
@@ -452,17 +457,23 @@ class _SimpleDependency(_RuntimeDependency):
                 return True
 
 
-    def on_service_modify(self, svc_ref):
+    def on_service_modify(self, svc_ref, old_properties):
         """
         Called when a service has been modified in the framework
         
         :param svc_ref: A service reference
+        :param old_properties: Previous properties values
         """
         with self._lock:
             if self.reference is None:
                 # A previously registered service now matches our filter
                 return self.on_service_arrival(svc_ref)
 
+            else:
+                # Notify the property modification
+                self._ipopo_instance.update(self, self._value, self.reference,
+                                            old_properties)
+
 
     def stop(self):
         """
@@ -614,11 +625,12 @@ class _AggregateDependency(_RuntimeDependency):
                 return True
 
 
-    def on_service_modify(self, svc_ref):
+    def on_service_modify(self, svc_ref, old_properties):
         """
         Called when a service has been modified in the framework
         
         :param svc_ref: A service reference
+        :param old_properties: Previous properties values
         :return: A tuple (added, (service, reference)) if the dependency has
                  been changed, else None
         """
@@ -627,6 +639,11 @@ class _AggregateDependency(_RuntimeDependency):
                 # A previously registered service now matches our filter
                 return self.on_service_arrival(svc_ref)
 
+            else:
+                # Notify the property modification
+                self._ipopo_instance.update(self, self.services[svc_ref],
+                                            svc_ref, old_properties)
+
 
     def stop(self):
         """
@@ -839,8 +856,8 @@ class FactoryContext(object):
     Represents the data stored in a component factory (class)
     """
 
-    __basic_fields = ('callbacks', 'name', 'properties', 'properties_fields',
-                      'provides')
+    __basic_fields = ('callbacks', 'field_callbacks', 'name', 'properties',
+                      'properties_fields', 'provides')
 
     def __init__(self):
         """
@@ -852,6 +869,9 @@ class FactoryContext(object):
         # Callbacks : Kind -> callback method
         self.callbacks = {}
 
+        # Field callbacks: Field -> {Kind -> Callback}
+        self.field_callbacks = {}
+
         # The factory name
         self.name = None
 
@@ -882,8 +902,8 @@ class FactoryContext(object):
             return False
 
         # Do not compare the bundle context, as it must not be stored
-        for field in ('name', 'callbacks', 'properties', 'properties_fields',
-                      'requirements'):
+        for field in ('name', 'callbacks', 'field_callbacks', 'properties',
+                      'properties_fields', 'requirements'):
             # Comparable fields
             if getattr(self, field, None) != getattr(other, field, None):
                 return False
@@ -913,8 +933,8 @@ class FactoryContext(object):
         context = FactoryContext()
 
         direct = ("bundle_context", "name")
-        copied = ("callbacks", "properties", "properties_fields",
-                  "requirements")
+        copied = ("callbacks", "field_callbacks", "properties",
+                  "properties_fields", "requirements")
         lists = ("provides",)
 
         # Direct copy of primitive values
@@ -1088,6 +1108,19 @@ class ComponentContext(object):
         return self.factory_context.callbacks.get(event, None)
 
 
+    def get_field_callback(self, field, event):
+        """
+        Retrieves the registered method for the given event. Returns None if not
+        found
+
+        :param field: Name of the dependency field
+        :param event: A component life cycle event
+        :return: The callback associated to the given event
+        """
+        return self.factory_context.field_callbacks.get(field,
+                                                        {}).get(event, None)
+
+
     def get_factory_name(self):
         """
         Retrieves the component factory name
@@ -1229,6 +1262,16 @@ class _StoredInstance(object):
             self.check_lifecycle()
 
 
+    def update(self, dependency, svc, svc_ref, old_properties):
+        """
+        Called by a dependency manager when the properties of an injected
+        dependency have been updated.
+        """
+        with self._lock:
+            self.__update_binding(dependency, svc, svc_ref, old_properties)
+            self.check_lifecycle()
+
+
     def unbind(self, dependency, svc, svc_ref):
         """
         Called by a dependency manager to remove an injected service and to
@@ -1511,6 +1554,30 @@ class _StoredInstance(object):
             return result
 
 
+    def __field_callback(self, field, event, *args, **kwargs):
+        """
+        Calls the registered method in the component for the given field event
+
+        :param field: A field name
+        :param event: An event (IPOPO_CALLBACK_VALIDATE, ...)
+        :return: The callback result, or None
+        :raise Exception: Something went wrong
+        """
+        with self._lock:
+            comp_callback = self.context.get_field_callback(field, event)
+            if not comp_callback:
+                # No registered callback
+                return True
+
+            # Call it
+            result = comp_callback(self.instance, field, *args, **kwargs)
+            if result is None:
+                # Special case, if the call back returns nothing
+                return True
+
+            return result
+
+
     def __safe_callback(self, event, *args, **kwargs):
         """
         Calls the registered method in the component for the given event,
@@ -1549,6 +1616,45 @@ class _StoredInstance(object):
                 return False
 
 
+    def __safe_field_callback(self, field, event, *args, **kwargs):
+        """
+        Calls the registered method in the component for the given event,
+        ignoring raised exceptions
+
+        :param field: Name of the modified field
+        :param event: A field event (IPOPO_CALLBACK_BIND_FIELD, ...)
+        :return: The callback result, or None
+        """
+        with self._lock:
+            if self.state == _StoredInstance.KILLED:
+                # Invalid state
+                return None
+
+            try:
+                return self.__field_callback(field, event, *args, **kwargs)
+
+            except pelix.FrameworkException as ex:
+                # Important error
+                _logger.exception("Critical error calling back %s: %s",
+                                  self.name, ex)
+
+                # Kill the component
+                self._ipopo_service.kill(self.name)
+
+                if ex.needs_stop:
+                    # Framework must be stopped...
+                    _logger.error("%s said that the Framework must be stopped.",
+                                  self.name)
+                    self.bundle_context.get_bundle(0).stop()
+                return False
+
+            except:
+                _logger.exception("Component '{0}' : error calling callback " \
+                                  "method for event {1}".format(self.name,
+                                                                event))
+                return False
+
+
     def __safe_handler_callback(self, handler, method_name, *args, **kwargs):
         """
         Calls the given method with the given arguments in the given handler.
@@ -1674,6 +1780,30 @@ class _StoredInstance(object):
             self.__safe_callback(constants.IPOPO_CALLBACK_BIND,
                                  service, reference)
 
+            self.__safe_field_callback(dependency.field,
+                                       constants.IPOPO_CALLBACK_BIND_FIELD,
+                                       service, reference)
+
+
+    def __update_binding(self, dependency, service, reference, old_properties):
+        """
+        Calls back component binding and field binding methods when the
+        properties of an injected dependency have been updated.
+        
+        :param dependency: The dependency handler
+        :param service: The injected service
+        :param reference: The reference of the injected service
+        :param old_properties: Previous properties of the dependency
+        """
+        with self._lock:
+            # Call the component back
+            self.__safe_field_callback(dependency.field,
+                                       constants.IPOPO_CALLBACK_UPDATE_FIELD,
+                                       service, reference, old_properties)
+
+            self.__safe_callback(constants.IPOPO_CALLBACK_UPDATE,
+                                 service, reference, old_properties)
+
 
     def __unset_binding(self, dependency, service, reference):
         """
@@ -1685,6 +1815,10 @@ class _StoredInstance(object):
         """
         with self._lock:
             # Call the component back
+            self.__safe_field_callback(dependency.field,
+                                       constants.IPOPO_CALLBACK_UNBIND_FIELD,
+                                       service, reference)
+
             self.__safe_callback(constants.IPOPO_CALLBACK_UNBIND,
                                  service, reference)
 
@@ -1878,6 +2012,13 @@ def _manipulate_component(instance, stored_instance):
         # Avoid injection of unused instance fields...
         controllers = set([value[1] for value in provides_tuples if value[1]])
         if controllers:
+            # Controllers are valid by default
+            for name in controllers:
+                # Get the current value of the member (True by default)
+                controller_value = getattr(instance, name, True)
+                # Store the controller value
+                stored_instance.set_controller_state(name, controller_value)
+
             # Prepare the methods names
             getter_name = "{0}{1}".format(constants.IPOPO_CONTROLLER_PREFIX,
                                           constants.IPOPO_GETTER_SUFFIX)
@@ -1889,10 +2030,6 @@ def _manipulate_component(instance, stored_instance):
             setattr(instance, getter_name, getter)
             setattr(instance, setter_name, setter)
 
-            # Controllers are valid by default
-            for name in controllers:
-                stored_instance.set_controller_state(name, True)
-
 
 # ------------------------------------------------------------------------------
 
@@ -1919,6 +2056,9 @@ class _IPopoService(object):
         # Event listeners
         self.__listeners = []
 
+        # Auto-restarted components (Bundle -> [(factory, name, properties)]
+        self.__auto_restart = {}
+
         # Service state
         self.running = False
 
@@ -1942,6 +2082,69 @@ class _IPopoService(object):
                     if stored_instance.factory_name == factory_name]
 
 
+    def _autorestart_store_components(self, bundle):
+        """
+        Stores the components of the given bundle with the auto-restart property
+        
+        :param bundle: A Bundle object
+        """
+        with self.__instances_lock:
+            # Prepare the list of components
+            store = self.__auto_restart.setdefault(bundle, [])
+
+            for stored_instance in self.__instances.values():
+                # Get the factory name
+                factory = stored_instance.factory_name
+
+                if self.get_factory_bundle(factory) is bundle:
+                    # Factory from this bundle
+
+                    # Test component properties
+                    properties = stored_instance.context.properties
+                    if properties.get(constants.IPOPO_AUTO_RESTART):
+                        # Auto-restart property found
+                        store.append((factory, stored_instance.name,
+                                      properties))
+
+
+    def _autorestart_components(self, bundle):
+        """
+        Restart the components of the given bundle
+        
+        :param bundle: A Bundle object
+        """
+        with self.__instances_lock:
+
+            instances = self.__auto_restart.get(bundle)
+            if not instances:
+                # Nothing to do
+                return
+
+            for factory, name, properties in instances:
+                try:
+                    # Instantiate the given component
+                    self.instantiate(factory, name, properties)
+
+                except Exception as ex:
+                    # Log error, but continue to work
+                    _logger.exception("Error restarting component '%s' ('%s')"
+                                      "from bundle %s (%d): %s", name, factory,
+                                      bundle.get_symbolic_name(),
+                                      bundle.get_bundle_id(), ex)
+
+
+    def _autorestart_clear_components(self, bundle):
+        """
+        Clear the list of auto-restart components of the given bundle
+        
+        :param bundle: A Bundle object
+        """
+        with self.__instances_lock:
+            # Simply delete the entry
+            if bundle in self.__auto_restart:
+                del self.__auto_restart[bundle]
+
+
     def _fire_ipopo_event(self, kind, factory_name, component_name=None):
         """
         Triggers an iPOPO event
@@ -2509,6 +2712,19 @@ class _IPopoActivator(object):
             # factories.
             self.service._register_bundle_factories(bundle)
 
+        elif kind == BundleEvent.UPDATE_BEGIN:
+            # A bundle will be updated, store its auto-restart component
+            self.service._autorestart_store_components(bundle)
+
+        elif kind == BundleEvent.UPDATED:
+            # Update has finished, restart stored components
+            self.service._autorestart_components(bundle)
+            self.service._autorestart_clear_components(bundle)
+
+        elif kind == BundleEvent.UPDATE_FAILED:
+            # Update failed, clean the stored components
+            self.service._autorestart_clear_components(bundle)
+
 
     def framework_stopping(self):
         """
